package org.tbasket.endpoint

import io.getquill.context.qzio.{ZioContext, ZioJdbcContext}
import io.getquill.context.sql.idiom.SqlIdiom
import io.getquill.idiom.Idiom
import io.getquill.{Literal, NamingStrategy, SqliteDialect}
import org.apache.logging.log4j.LogManager
import org.tbasket.auth.Authenticator
import org.tbasket.data.DatabaseContext
import org.tbasket.endpoint.Endpoint.LOG
import org.tbasket.handler.LoginPageHandler
import zio.*
import zio.http.*
import zio.http.ServerConfig.LeakDetectionLevel
import zio.http.model.Method.{GET, POST}
import zio.http.model.Status
import zio.http.model.Status.InternalServerError
import zio.http.netty.client.ConnectionPool

import javax.sql.DataSource
import scala.collection.mutable

class Endpoint(port: Int):
  
  
  // set generic required headers
  private def applyGenerics(response: Response): Response =
    response.withAccessControlAllowOrigin("*")
  
  private val app  = Http.collectZIO[Request] {
    case r@POST -> _ / "login" =>
      LoginPageHandler.post(r)
      
    case r@method -> path =>
      val ipInsights = r.remoteAddress
        .map(ip => s": request received from $ip.")
        .getOrElse("")
      LOG.error(
        s"Was unable to find a handler for request '$path' with method $method ${ipInsights}"
      )
      ZIO.succeed(Response(Status.NotFound))
  }.catchAllCause(handleUnexpectedError)
    .map(applyGenerics)
  
  private def handleUnexpectedError(cause: Cause[Throwable]): HttpApp[Any, Throwable] = {
    def report(kind: String, value: Throwable = null, trace: StackTrace = StackTrace.none) =
      LOG.error(s"Received unhandled $kind cause ${if value == null then "" else ": " + value}")
      LOG.error(trace)
    
    cause match
      case Cause.Empty                 => report("empty")
      case Cause.Fail(e, trace)        => report("failure", e, trace)
      case Cause.Die(e, trace)         => report("die", e, trace)
      case Cause.Interrupt(fiberId, e) => report(s"interruption of $fiberId", null, e)
      case Cause.Stackless(cause, _)   =>
        LOG.error("stackless error :")
        handleUnexpectedError(cause)
      case Cause.Then(left, right)     =>
        handleUnexpectedError(left)
        LOG.error("**THEN this error occurred : **")
        handleUnexpectedError(right)
      case Cause.Both(left, right)     =>
        handleUnexpectedError(left)
        LOG.error("**AND this error also occurred (async) : **")
        handleUnexpectedError(right)
    
    
    Http.succeed(Response.status(InternalServerError))
  }
  
  val run =
    val config = ServerConfig.default
      .port(port)
      .leakDetection(LeakDetectionLevel.PARANOID)
    
    val serverConfigLayer = ServerConfig.live(config)
    Server.install(app).flatMap { port =>
      LOG.info(s"Listening API entries on $port")
      ZIO.never
    }.provideSome[DatabaseContext & Authenticator & DataSource](
      Scope.default,
      serverConfigLayer,
      ConnectionPool.fixed(4),
      ClientConfig.default,
      Server.live,
      Client.live
    )

object Endpoint:
  final val LOG = LogManager.getLogger("API")
